%{
/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <libfawk/config.h>
#include <stdlib.h>
#include <libfawk/libfawk.h>
#include <libfawk/parser.h>
#include "fawk.y.tab.h"
#include "parser_fawk.h"
#include <libfawk/symtab.h>
#include <libfawk/compile.h>
#include <libfawk/execute.h>
%}

%union {
	char *str;
	fawk_num_t num;
}

%token T_FUNCTION T_EQ T_NEQ T_GTEQ T_LTEQ
%token T_PLEQ T_MIEQ T_MUEQ T_DIEQ T_MOEQ T_NIL
%token T_IF T_ELSE T_FOR T_IN T_DO T_WHILE T_RETURN

%token <str> T_ID T_STRING
%token <num> T_NUM

%right '=' T_PLEQ T_MIEQ T_MUEQ T_DIEQ T_MOEQ
%right '?' ':'
%left T_OR
%left T_AND
%left T_IN
%left T_EQ T_NEQ  T_LTEQ '<' T_GTEQ '>'
%left '@'
%left '+' '-'
%left '*' '/' '%'
%left '!'
%left T_MM T_PP
%right '(' ')'
%%

script:
	{ fawkc_addi(ctx, FAWKI_ABORT); fawkc_addcs(ctx, "uninitialized execute"); }
	program
	{ fawkc_addi(ctx, FAWKI_ABORT); fawkc_addcs(ctx, "ran beyond the script"); }
	;

program:
	func_def program
	| /* empty */
	;

func_def:
	T_FUNCTION T_ID                              { ctx->compiler.numfixedargs = -1; ctx->compiler.numargs = 0; }
		'(' arg_def ')'                            { fawk_symtab_regfunc(ctx, $2, FAWK_CURR_IP(), ctx->compiler.numargs, ctx->compiler.numfixedargs); fawk_free(ctx, $2); ctx->fp = ctx->sp; }
		'{' statements '}'
		{
			int n;
			fawkc_addi(ctx, FAWKI_PUSH_NIL);
			fawkc_addi(ctx, FAWKI_RET);
			fawkc_addnum(ctx, ctx->compiler.numargs + (ctx->compiler.numfixedargs >= 0));
			for(n = 0; n < ctx->compiler.numargs + (ctx->compiler.numfixedargs >= 0); n++) {
				fawk_cell_t cell;
				fawk_pop(ctx, &cell);
				fawk_cell_free(ctx, &cell);
			}
			ctx->fp = 0;
		}
	;

arg_def:
	  /* empty */
	| T_ID                                       { fawk_push_str(ctx, $1); fawk_free(ctx, $1); ctx->compiler.numargs++; }
	| '.' '.' '.'                                { ctx->compiler.numfixedargs = ctx->compiler.numargs; fawk_push_str(ctx, "VARARG"); }
	| arg_def ',' T_ID                           { fawk_push_str(ctx, $3); fawk_free(ctx, $3); ctx->compiler.numargs++; }
	| arg_def ',' '.' '.' '.'                    { ctx->compiler.numfixedargs = ctx->compiler.numargs; fawk_push_str(ctx, "VARARG"); }
	;

statement:
	  expr ';'           { fawkc_addi(ctx, FAWKI_POP); /* void experssion's result value shall be removed from the stack */ }
	| s_if
	| s_for
	| s_while
	| s_do
	| T_RETURN expr ';'  { fawkc_addi(ctx, FAWKI_RET); fawkc_addnum(ctx, ctx->compiler.numargs); }
	| T_RETURN ';'       { fawkc_addi(ctx, FAWKI_PUSH_NIL); fawkc_addi(ctx, FAWKI_RET); fawkc_addnum(ctx, ctx->compiler.numargs); }
	| '{' statements '}'
	| ';'
	;

statements:
	  /* empty */
	| statements statement
	;

s_if:
		if_then
			{
				size_t jmp1 = fawk_pop_num(ctx, 1);
				ctx->code.code[jmp1].data.num = FAWK_CURR_IP();
			}
		| if_then T_ELSE
		{
			fawkc_addi(ctx, FAWKI_JMP);
			FAWK_PUSH_IP(); fawkc_addnum(ctx, 777);
		}
		statement
		{
			size_t jmp_then_post = fawk_pop_num(ctx, 1), jmp_if = fawk_pop_num(ctx, 1);
			ctx->code.code[jmp_then_post].data.num = FAWK_CURR_IP();
			ctx->code.code[jmp_if].data.num = jmp_then_post+1;
		}
	;

if_then:
	T_IF '(' expr ')'    { fawkc_addi(ctx, FAWKI_POPJZ); FAWK_PUSH_IP(); fawkc_addnum(ctx, 777); }
		statement
	;

s_for:
	  s_for_plain
	| s_for_in
	;

s_for_in:
	T_FOR '(' var T_IN var ')'
			{
				fawkc_addi(ctx, FAWKI_FORIN_FIRST);
				fawkc_addi(ctx, FAWKI_JMP); /* skip statements first, so next can load the variable first */
				fawkc_addnum(ctx, 0);
				FAWK_PUSH_IP();
			}
		statement
			{
				size_t begin = fawk_pop_num(ctx, 1);
				fawkc_addi(ctx, FAWKI_FORIN_NEXT);
				fawkc_addnum(ctx, begin);
				ctx->code.code[begin-1].data.num = FAWK_CURR_IP()-2;
			}
	;


s_for_plain:
	T_FOR '(' maybe_expr ';'
			{
				fawkc_addi(ctx, FAWKI_POP); /* void expr */
				FAWK_PUSH_IP(); /* slot A */
			}
		maybe_expr ';'
			{
				fawkc_addi(ctx, FAWKI_POPJZ); /* jump out on false */
				FAWK_PUSH_IP(); /* slot B */
				fawkc_addnum(ctx, 0);
				fawkc_addi(ctx, FAWKI_JMP); /* skip 3rd expr */
				FAWK_PUSH_IP(); /* slot C */
				fawkc_addnum(ctx, 0);
			}
		maybe_expr ')'    /* 3rd expr - shall be run at the end */
			{
				fawkc_addi(ctx, FAWKI_POP); /* void expr */
				fawkc_addi(ctx, FAWKI_JMP); /* jump back to the check */
				FAWK_PUSH_IP(); /* slot D */
				fawkc_addnum(ctx, 0);
			}
		statement
			{
				size_t jback3rd, jskip, jout, rerun;

				jback3rd = fawk_pop_num(ctx, 1); /* slot D */
				jskip = fawk_pop_num(ctx, 1);    /* slot C */
				jout = fawk_pop_num(ctx, 1);     /* slot B */
				rerun = fawk_pop_num(ctx, 1);    /* slot A */

				fawkc_addi(ctx, FAWKI_JMP); /* jump back to execute the 3rd expr */
				fawkc_addnum(ctx, jskip+1);

				ctx->code.code[jback3rd].data.num = rerun;
				ctx->code.code[jskip].data.num = jback3rd+1;
				ctx->code.code[jout].data.num = FAWK_CURR_IP();
			}
	;


s_do:
	T_DO                                            { FAWK_PUSH_IP(); }
		statement T_WHILE '(' expr ')'                { fawkc_addi(ctx, FAWKI_POPJNZ); fawkc_addnum(ctx, fawk_pop_num(ctx, 1)); }
	;

s_while:
	T_WHILE                          { FAWK_PUSH_IP(); }
		'(' expr ')'                   { fawkc_addi(ctx, FAWKI_POPJZ); FAWK_PUSH_IP(); fawkc_addnum(ctx, 0); }
		statement                      { loop_pretest_jumpback(); }
	;

expr:
	  T_NUM                       { fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, $1); }
	| T_STRING                    { fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $1); fawk_free(ctx, $1); }
	| T_NIL                       { fawkc_addi(ctx, FAWKI_PUSH_NIL); }
	| var                         { fawkc_addi(ctx, FAWKI_PUSH_SYMVAL); }
	| '&' var                     /* done in var */
	| e_call
	| e_assign
	| '(' expr ')'                /* done implicitly by the parser */
	| expr T_EQ expr              { fawkc_addi(ctx, FAWKI_EQ); }
	| expr T_NEQ expr             { fawkc_addi(ctx, FAWKI_NEQ); }
	| expr T_GTEQ expr            { fawkc_addi(ctx, FAWKI_GTEQ); }
	| expr T_LTEQ expr            { fawkc_addi(ctx, FAWKI_LTEQ); }
	| expr '>' expr               { fawkc_addi(ctx, FAWKI_GT); }
	| expr '<' expr               { fawkc_addi(ctx, FAWKI_LT); }
	| expr T_IN var               { fawkc_addi(ctx, FAWKI_IN); }
	| e_trenary
	| e_or
	| e_and
	| expr '+' expr               { fawkc_addi(ctx, FAWKI_ADD); }
	| expr '-' expr               { fawkc_addi(ctx, FAWKI_SUB); }
	| expr '*' expr               { fawkc_addi(ctx, FAWKI_MUL); }
	| expr '/' expr               { fawkc_addi(ctx, FAWKI_DIV); }
	| expr '%' expr               { fawkc_addi(ctx, FAWKI_MOD); }
	| '-' expr                    { fawkc_addi(ctx, FAWKI_NEG); }
	| '+' expr                    /* a positive number is the same number, don't do anything */
	| '!' expr                    { fawkc_addi(ctx, FAWKI_NOT); }
	| expr '@' expr               { fawkc_addi(ctx, FAWKI_CONCAT); }
	| e_inc_dec
	;

maybe_expr:
	  expr
	| /* empty */                 { fawkc_addi(ctx, FAWKI_PUSH_NUM); fawkc_addnum(ctx, 1); }
	;

var:
	T_ID          { fawk_push_num(ctx, ctx->compiler.numidx); ctx->compiler.numidx = 0; }
		arridx
		{
			fawkc_addi(ctx, FAWKI_MAKE_SYMREF);
			fawkc_addsymref(ctx, $1, ctx->compiler.numidx, 0);
			fawkc_addnum(ctx, ctx->compiler.numidx);
			fawk_free(ctx, $1);
			ctx->compiler.numidx = fawk_pop_num(ctx, 1);
		}
	;

e_inc_dec:
	  var T_PP   { fawkc_addi(ctx, FAWKI_INCDEC); fawkc_addnum(ctx, FAWK_INCDEC_INC | FAWK_INCDEC_POST); }
	| var T_MM   { fawkc_addi(ctx, FAWKI_INCDEC); fawkc_addnum(ctx, FAWK_INCDEC_POST); }
	| T_PP var   { fawkc_addi(ctx, FAWKI_INCDEC); fawkc_addnum(ctx, FAWK_INCDEC_INC); }
	| T_MM var   { fawkc_addi(ctx, FAWKI_INCDEC); fawkc_addnum(ctx, 0); }
	;

arridx:
	  /* empty */
	| arridx '[' ']'           { ctx->compiler.numidx = -1; }
	| arridx '[' aix_expr ']'  { ctx->compiler.numidx++; }
	| arridx '.' T_ID          { ctx->compiler.numidx++; fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $3); fawk_free(ctx, $3); }
	| arridx '.' T_STRING      { ctx->compiler.numidx++; fawkc_addi(ctx, FAWKI_PUSH_STR); fawkc_adds(ctx, $3); fawk_free(ctx, $3); }
	;

aix_expr: /* array index expression */
	  expr
	| aix_expr ','  { parse_aix_expr(); }
		expr          { fawkc_addi(ctx, FAWKI_CONCAT); }
	;

e_call:
	expr '('                      { fawk_push_num(ctx, ctx->compiler.numargs); ctx->compiler.numargs = 0; ctx->code.used--; /* remove push_symval on the func symref */ }
		call_arg ')'
		{
			size_t old_numargs = fawk_pop_num(ctx, 1);
			fawkc_addi(ctx, FAWKI_CALL);
			fawkc_addnum(ctx, ctx->compiler.numargs);
			ctx->compiler.numargs = old_numargs;
		}
	;

call_arg:
	  /* empty */
	| expr                 { ctx->compiler.numargs++; }
	| expr ',' call_arg    { ctx->compiler.numargs++; }
	;

e_trenary:
	expr '?'  { fawkc_addi(ctx, FAWKI_POPJZ); FAWK_PUSH_IP(); fawkc_addnum(ctx, 777); }
	expr ':'  { fawkc_addi(ctx, FAWKI_JMP); FAWK_PUSH_IP(); fawkc_addnum(ctx, 888); }
	expr
		{
			size_t jmp1, jmp2;
			jmp2 = fawk_pop_num(ctx, 1); jmp1 = fawk_pop_num(ctx, 1);
			ctx->code.code[jmp1].data.num = jmp2+1;
			ctx->code.code[jmp2].data.num = FAWK_CURR_IP();
		}
	;

/* lazy OR evaluation */
e_or:
	expr T_OR             { lazy_binop1(ctx, 1); }
		expr                { lazy_binop2(ctx, 1); }
	;

/* lazy AND evaluation */
e_and:
	expr T_AND             { lazy_binop1(ctx, 0); }
		expr                 { lazy_binop2(ctx, 0); }
	;

e_assign:
	  var '=' expr     { fawkc_addi(ctx, FAWKI_SET); fawkc_addi(ctx, FAWKI_PUSH_SYMVAL); }
	| var T_PLEQ { fawkc_addi(ctx, FAWKI_PUSH_TOPVAR); } expr  { fawkc_addi(ctx, FAWKI_ADD); fawkc_addi(ctx, FAWKI_SET); }
	| var T_MIEQ { fawkc_addi(ctx, FAWKI_PUSH_TOPVAR); } expr  { fawkc_addi(ctx, FAWKI_SUB); fawkc_addi(ctx, FAWKI_SET); }
	| var T_MUEQ { fawkc_addi(ctx, FAWKI_PUSH_TOPVAR); } expr  { fawkc_addi(ctx, FAWKI_MUL); fawkc_addi(ctx, FAWKI_SET); }
	| var T_DIEQ { fawkc_addi(ctx, FAWKI_PUSH_TOPVAR); } expr  { fawkc_addi(ctx, FAWKI_DIV); fawkc_addi(ctx, FAWKI_SET); }
	| var T_MOEQ { fawkc_addi(ctx, FAWKI_PUSH_TOPVAR); } expr  { fawkc_addi(ctx, FAWKI_MOD); fawkc_addi(ctx, FAWKI_SET); }
	;

%%

FAWK_API void fawk_error(fawk_ctx_t *ctx, fawk_STYPE tok, const char *msg) { libfawk_error(ctx, msg, ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1); }

FAWK_API int fawk_lex_fawk(fawk_STYPE *lval, fawk_ctx_t *ctx);
FAWK_API int fawk_parse_fawk(fawk_ctx_t *ctx) { fawk_parser_loop(fawk_yyctx_t, fawk_STYPE, fawk_lex_fawk, fawk_parse, ctx, fawk_RES_NEXT, fawk_RES_DONE); }
